@@ -1,8 +1,7 @@
-from collections import Counter
 import math
-from random import choice, randint
 import string
-
+from collections import Counter
+from random import choice, getrandbits, randint, random
 from timeit import default_timer
 from typing import Any, Callable, Dict, List, Tuple
 
@@ -23,6 +22,9 @@ class bigO:
 
     runtime(function, array, size) -> Tuple[executionTime, sorted result]:
         Returns executionTime and the result
+        
+    compare(function1, function2, array, size) -> Dict{functionName: executionTime}
+        Returns dictionary with execution time on each function
 
     Usage
     -----
@@ -34,6 +36,7 @@ class bigO:
         lib.test(mySort, "random")
         lib.test_all(mySort)
         lib.runtime(algorithm.bubbleSort, "random", 5000)
+        lib.compare(algorithm.bubbleSort, algorithm.insertSort, "random", 5000)
     """
 
     def __init__(self):
@@ -151,33 +154,47 @@ class bigO:
 
         return bestFit
 
-    def genRandomArray(_, size: int):
-        array = [randint(-size, size) for _ in range(size)]
+    def genRandomPositive(_, size: int = 10):
+        return [randint(1, size) for i in range(size)]
+
+    def genRandomArray(_, size: int = 10):
+        return [randint(-size, size) for i in range(size)]
+
+    def genRandomBigArray(_, size: int = 10):
+        array = []
+        for _ in range(size):
+            isPositive = random() < 0.5
+            nextValue = getrandbits(50)  # More than 100 trillion
+            if not isPositive:
+                nextValue = -nextValue
+            array.append(nextValue)
         return array
 
-    def genRandomString(_, size: int):
+    def genRandomString(_, stringLen: int = None, size: int = 10):
+        if stringLen == None:
+            stringLen = size // 2
+
         letters = string.ascii_lowercase + string.digits
         array = [
-            "".join(choice(letters) for _ in range(size)) for _ in range(size)
+            "".join(choice(letters) for i in range(randint(1, stringLen)))
+            for j in range(size)
         ]  # secrets.choice?
         return array
 
-    def genSortedArray(_, size: int):
-        array = [i for i in range(size)]
-        return array
+    def genSortedArray(_, size: int = 10):
+        return [i for i in range(size)]
 
-    def genReversedArray(_, size: int):
-        array = [i for i in reversed(range(size))]
-        return array
+    def genReversedArray(_, size: int = 10):
+        return [i for i in reversed(range(size))]
 
-    def genPartialArray(self, size: int):
+    def genPartialArray(self, size: int = 10):
         array = self.genRandomArray(size)
         sorted_array = self.genSortedArray(size)
 
         array[size // 4 : size // 2] = sorted_array[size // 4 : size // 2]
         return array
 
-    def genKsortedArray(self, size: int, k: int = None):
+    def genKsortedArray(self, size: int = 10, k: int = None):
         def reverseRange(array, a, b):
             i = a
             j = b - 1
@@ -208,6 +225,36 @@ class bigO:
 
         return array
 
+    def genEqualArray(_, size: int = 10):
+        n = randint(-size, size)
+        return [n for i in range(size)]
+
+    def genAlmostEqualArray(_, size: int = 10):
+        return [randint(-1, 1) + size for i in range(size)]
+
+    def genHoleArray(
+        self, size: int = 10
+    ):  # returns equal array with only one different element
+        arr = self.genEqualArray(size)
+        arr[randint(-size, size)] = -9999
+        return arr
+
+    def isAscendingSorted(_, array: List[Any]):
+        """Is correctly ascending sorted? Time: O(n)
+
+        Args:
+            array [List]: Array to check if it is sorted correctly
+
+        Returns:
+            isSorted, index [bool, int]: returns True/False with unsorted index
+        """
+        # Ascending order
+        for i in range(len(array) - 1):
+            if array[i] > array[i + 1]:
+                return False, i + 1
+
+        return True, None
+
     def test(
         self,
         functionName: Callable,
@@ -220,7 +267,8 @@ class bigO:
 
         Args:
             functionName (Callable): a function to call |
-            array (str): "random", "sorted", "reversed", "partial", "Ksorted", "string" |
+            array (str): "random", "big", "sorted", "reversed", "partial", "Ksorted", "string",
+            "hole", "equal", "almost_equal" |
             limit (bool): To terminate before it takes forever to sort (usually 10,000) |
             prtResult (bool): Whether to print the result by itself (default = True)
 
@@ -255,6 +303,8 @@ class bigO:
             array = array.lower()
             if array == "random":
                 nums = self.genRandomArray(size)
+            elif array == "big":
+                nums = self.genRandomBigArray(size)
             elif array == "sorted":
                 nums = self.genSortedArray(size)
             elif array == "partial":
@@ -264,7 +314,13 @@ class bigO:
             elif array == "ksorted":
                 nums = self.genKsortedArray(size, size.bit_length())
             elif array == "string":
-                nums = self.genRandomString(size)
+                nums = self.genRandomString(stringLen=100, size=size)
+            elif array == "hole":
+                nums = self.genHoleArray(size)
+            elif array == "equal":
+                nums = self.genEqualArray(size)
+            elif array == "almost_equal":
+                nums = self.genAlmostEqualArray(size)
             # elif array == "custom":
             #    nums = custom
             #    assert len(nums) != 0, "Please, pass the custom array you want.
@@ -281,13 +337,20 @@ class bigO:
                 currentIter += 1
 
                 if result != None:
-                    assert result == sorted(
-                        nums
-                    ), "This function doesn't sort correctly."
+                    isSorted, index = self.isAscendingSorted(result)
+                    if index == 1:
+                        msg = f"{result[index - 1]}, {result[index]}..."
+                    elif index == len(result) - 1:
+                        msg = f"...{result[index - 1]}, {result[index]}"
+                    elif isinstance(index, int):
+                        msg = f"...{result[index - 1]}, {result[index]}, {result[index + 1]}..."
+                    assert (
+                        isSorted
+                    ), f"{functionName.__name__} doesn't sort correctly.\nAt {index} index: [{msg}]"
 
             if (
-                timeTaken >= 5 and limit
-            ):  # if it takes more than 5 seconds to sort one array, break
+                timeTaken >= 4 and limit
+            ):  # if it takes more than 4 seconds to sort an array, then break it
                 isSlow = True
 
             timeTaken /= maxIter
@@ -319,7 +382,14 @@ class bigO:
         Returns:
             Dict[str, int]: ex) {"random": "O(n)" ...}
         """
-        result = {"random": 0, "sorted": 0, "reversed": 0, "partial": 0, "Ksorted": 0}
+        result = {
+            "random": 0,
+            "sorted": 0,
+            "reversed": 0,
+            "partial": 0,
+            "Ksorted": 0,
+            "almost_equal": 0,
+        }
 
         bestCase = self.complexity2int("O(n^3)")
         worstCase = self.complexity2int("O(1)")
@@ -356,7 +426,8 @@ class bigO:
 
         Args:
             function [Callable]: a function to call |
-            array: "random", "sorted", "partial", "reversed", "Ksorted" or your custom array |
+            array: "random", "big", "sorted", "partial", "reversed", "Ksorted" ,
+            "hole", "equal", "almost_equal" or your custom array |
             size [int]: How big test array should be |
             epoch [int]: How many tests to run and calculte average |
             prtResult (bool): Whether to print the result by itself (default = True) |
@@ -375,6 +446,8 @@ class bigO:
             array = array.lower()
             if array == "random":
                 nums = self.genRandomArray(size)
+            elif array == "big":
+                nums = self.genRandomBigArray(size)
             elif array == "sorted":
                 nums = self.genSortedArray(size)
             elif array == "partial":
@@ -385,6 +458,12 @@ class bigO:
                 nums = self.genKsortedArray(size, size.bit_length())
             elif array == "string":
                 nums = self.genRandomString(size)
+            elif array == "hole":
+                nums = self.genHoleArray(size)
+            elif array == "equal":
+                nums = self.genEqualArray(size)
+            elif array == "almost_equal":
+                nums = self.genAlmostEqualArray(size)
             else:  # default = random array
                 nums = self.genRandomArray(size)
 
@@ -401,9 +480,19 @@ class bigO:
             timeTaken += timeEnd - timeStart
 
             if result != None:
-                if result != sorted(nums):
+                isSorted, index = self.isAscendingSorted(result)
+                if index == 1:
+                    msg = f"{result[index - 1]}, {result[index]}..."
+                elif index == len(result) - 1:
+                    msg = f"...{result[index - 1]}, {result[index]}"
+                elif isinstance(index, int):
+                    msg = f"...{result[index - 1]}, {result[index]}, {result[index + 1]}..."
+
+                if not isSorted:
                     # Just see the result if it doesn't sort correctly
-                    print("This function doesn't sort correctly.")
+                    print(
+                        f"{function.__name__} doesn't sort correctly.\nAt {index} index: [{msg}]"
+                    )
 
         finalTime = timeTaken / epoch
 
@@ -413,7 +502,7 @@ class bigO:
         return finalTime, result
 
     def compare(
-        self, function1: Callable, function2: Callable, array: str, size: int
+        self, function1: Callable, function2: Callable, array, size: int = None
     ) -> Dict:
         """
         ex) compare(bubbleSort, insertSort, "random", 5000)
@@ -421,17 +510,32 @@ class bigO:
         Args:
             function1 [Callable]: a function to compare |
             function2 [Callable]: a function to compare |
-            array [str]: "random", "sorted", "partial", "reversed", "Ksorted", "all"
+            array [str]|[List]: "random", "big", "sorted", "partial", "reversed", "Ksorted", 
+            "hole", "equal", "almost_equal", "all" or your custom array |
             size [int]: How big test array should be |
 
         Returns:
             Dict: function1 execution time and function2 execution time
         """
+        s = ""
+
         if array == "all":
-            test = ["random", "sorted", "reversed", "partial", "Ksorted"]
+            test = [
+                "random",
+                "big",
+                "sorted",
+                "reversed",
+                "partial",
+                "Ksorted",
+                "hole",
+                "equal",
+                "almost_equal",
+            ]
             func1_sum = 0.0
             func2_sum = 0.0
-            print(f"Running {function1.__name__}(tests), {function2.__name__}(tests)")
+            wins = 0
+
+            print(f"Running {function1.__name__}(tests) vs {function2.__name__}(tests)")
             for arr in test:
                 function1_time, _ = self.runtime(
                     function1, arr, size, epoch=3, prtResult=False
@@ -443,11 +547,23 @@ class bigO:
                 )
                 func2_sum += function2_time
 
+                if function1_time > function2_time:
+                    wins += 1
+
             func1_sum /= len(test)
             func2_sum /= len(test)
             function1_time = func1_sum
             function2_time = func2_sum
+
+            wins = wins if function1_time > function2_time else len(test) - wins
+            array = f"{wins} of {len(test)}"
+            s = "s"
         else:
+            if isinstance(array, list):
+                nums = array
+                array = "custom"
+                size = len(nums)
+
             function1_time, _ = self.runtime(
                 function1, array, size, epoch=3, prtResult=False
             )
@@ -458,13 +574,15 @@ class bigO:
         timeDiff = abs(function1_time - function2_time)
 
         if function1_time < function2_time:
+            percentage = function2_time / function1_time * 100.0 - 100.0
             print(
-                f"{function1.__name__} is faster than {function2.__name__} on {array} case"
+                f"{function1.__name__} is {percentage:.1f}% faster than {function2.__name__} on {array} case{s}"
             )
             print(f"Time Difference: {timeDiff:.5f}s")
         else:
+            percentage = function1_time / function2_time * 100.0 - 100.0
             print(
-                f"{function2.__name__} is faster than {function1.__name__} on {array} case"
+                f"{function2.__name__} is {percentage:.1f}% faster than {function1.__name__} on {array} case{s}"
             )
             print(f"Time Difference: {timeDiff:.5f}s")
 
